#![no_std]
#![no_main]
#![allow(async_fn_in_trait)]
#![feature(impl_trait_in_assoc_type)]
#![feature(slice_split_once)]

use core::net::Ipv4Addr;
use core::slice;
use cyw43_pio::{DEFAULT_CLOCK_DIVIDER, PioSpi};
use dhcparse::dhcpv4::{DhcpOption, Encode as _, MessageType as DhcpMsgType};
use dhcparse::{dhcpv4, v4_options};
use embassy_executor::Spawner;
use embassy_net::udp::{PacketMetadata, UdpMetadata, UdpSocket};
use embassy_net::{
    Config, DhcpConfig, EthernetAddress, HardwareAddress, IpEndpoint, Ipv4Cidr, StackResources,
};
use embassy_rp::bind_interrupts;
use embassy_rp::clocks::RoscRng;
use embassy_rp::gpio::{Level, Output};
use embassy_rp::peripherals::USB;
use embassy_rp::peripherals::{DMA_CH0, PIO0};
use embassy_rp::pio::{InterruptHandler as PioInterruptHandler, Pio};
use embassy_rp::usb::{Driver, InterruptHandler as UsbInterruptHandler};
use embassy_time::{Instant, Timer};
use heapless::Vec;
use log::info;
use pico_website::unwrap;
use rand_core::RngCore;
use static_cell::StaticCell;
use {defmt_rtt as _, panic_probe as _};

mod game;
mod socket;

bind_interrupts!(struct Irqs {
    USBCTRL_IRQ => UsbInterruptHandler<USB>;
    PIO0_IRQ_0 => PioInterruptHandler<PIO0>;
});

#[embassy_executor::task]
async fn logger_task(driver: Driver<'static, USB>) {
    embassy_usb_logger::run!(1024, log::LevelFilter::Info, driver);
}

#[embassy_executor::task]
async fn cyw43_task(
    runner: cyw43::Runner<'static, Output<'static>, PioSpi<'static, PIO0, 0, DMA_CH0>>,
) -> ! {
    runner.run().await
}

#[embassy_executor::task]
async fn net_task(mut runner: embassy_net::Runner<'static, cyw43::NetDriver<'static>>) -> ! {
    runner.run().await
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    let p = embassy_rp::init(Default::default());
    let driver = Driver::new(p.USB, Irqs);
    spawner.spawn(logger_task(driver)).unwrap();
    let mut rng = RoscRng;

    let fw = include_bytes!("../cyw43-firmware/43439A0.bin");
    let clm = include_bytes!("../cyw43-firmware/43439A0_clm.bin");
    let pwr = Output::new(p.PIN_23, Level::Low);
    let cs = Output::new(p.PIN_25, Level::High);
    let mut pio = Pio::new(p.PIO0, Irqs);
    let spi = PioSpi::new(
        &mut pio.common,
        pio.sm0,
        DEFAULT_CLOCK_DIVIDER,
        pio.irq0,
        cs,
        p.PIN_24,
        p.PIN_29,
        p.DMA_CH0,
    );

    static STATE: StaticCell<cyw43::State> = StaticCell::new();
    let state = STATE.init(cyw43::State::new());
    let (net_device, mut control, runner) = cyw43::new(state, pwr, spi, fw).await;
    unwrap(spawner.spawn(cyw43_task(runner))).await;

    control.init(clm).await;
    control
        .set_power_management(cyw43::PowerManagementMode::PowerSave)
        .await;

    #[cfg(not(feature = "wifi-connect"))]
    // Use a link-local address for communication without DHCP server
    let config = Config::ipv4_static(embassy_net::StaticConfigV4 {
        address: embassy_net::Ipv4Cidr::new(embassy_net::Ipv4Address::new(192, 254, 0, 2), 24),
        dns_servers: heapless::Vec::new(),
        gateway: None,
    });

    #[cfg(feature = "wifi-connect")]
    let (wifi_conf, config) = {
        let wifi_conf = unwrap(serde_json_core::from_slice::<WifiConnectConf>(
            include_bytes!("../wifi.json"),
        ))
        .await
        .0;
        // Use a link-local address for communication without DHCP server
        // let config = Config::dhcpv4(embassy_net::DhcpConfig::default());
        let config = match wifi_conf.ip {
            Some(ip) => Config::ipv4_static(embassy_net::StaticConfigV4 {
                address: embassy_net::Ipv4Cidr::new(ip, 24),
                dns_servers: heapless::Vec::new(),
                gateway: None,
            }),
            None => Config::dhcpv4(DhcpConfig::default()),
        };
        (wifi_conf, config)
    };

    // Generate random seed
    let seed = rng.next_u64();

    // Init network stack
    static RESOURCES: StaticCell<StackResources<4>> = StaticCell::new();
    let (stack, runner) = embassy_net::new(
        net_device,
        config,
        RESOURCES.init(StackResources::new()),
        seed,
    );

    unwrap(spawner.spawn(net_task(runner))).await;

    #[cfg(not(feature = "wifi-connect"))]
    //control.start_ap_open("cyw43", 5).await;
    control.start_ap_wpa2("cyw43", "password", 5).await;

    #[cfg(feature = "wifi-connect")]
    {
        loop {
            match control
                .join(
                    wifi_conf.name,
                    cyw43::JoinOptions::new(wifi_conf.password.as_bytes()),
                )
                .await
            {
                Ok(_) => break,
                Err(err) => {
                    info!("join failed with status={}", err.status);
                }
            }
        }
        info!("Network joined!");
        info!("waiting for link...");
        stack.wait_link_up().await;
        // Wait for DHCP, not necessary when using static IP
        info!("waiting for DHCP...");
        stack.wait_config_up().await;
        // while !stack.is_config_up() {
        //     Timer::after_millis(100).await;
        // }
        info!("DHCP is now up!");
        info!(
            "ip : {}",
            unwrap(stack.config_v4().ok_or("no dhcp config"))
                .await
                .address
        )
    }

    let mut rx_buffer = [0; 4096];
    let mut tx_buffer = [0; 4096];
    let mut rx_meta = [PacketMetadata::EMPTY; 16];
    let mut tx_meta = [PacketMetadata::EMPTY; 16];
    let mut buf = [0; 4096];
    let mut res_buf = Vec::<u8, 4096>::new();
    loop {
        let mut socket = UdpSocket::new(
            stack,
            &mut rx_meta,
            &mut rx_buffer,
            &mut tx_meta,
            &mut tx_buffer,
        );
        unwrap(socket.bind(67)).await;

        loop {
            info!("receiving");
            let (n, ep) = unwrap(socket.recv_from(&mut buf).await).await;
            info!("received");
            Timer::after_secs(0).await;

            let msg = unwrap(dhcpv4::Message::new(&buf[..n])).await;

            let msg_type = unwrap(v4_options!(msg; MessageType required)).await;
            for o in unwrap(msg.options()).await {
                info!(
                    "{:?}",
                    match o {
                        Err(_) => break,
                        Ok(o) => o,
                    }
                );
            }
            info!("{:?}", msg_type);
            Timer::after_secs(0).await;

            match msg_type {
                DhcpMsgType::DISCOVER | DhcpMsgType::REQUEST => {
                    res_buf.clear();
                    res_buf.push(2).unwrap(); // op
                    res_buf.push(1).unwrap(); // htype
                    res_buf.push(6).unwrap(); // hlen
                    res_buf.push(0).unwrap(); // hops
                    res_buf.extend_from_slice(&buf[4..8]).unwrap(); // xid
                    res_buf.extend_from_slice(&[0; 2]).unwrap(); // secs
                    res_buf.extend_from_slice(&[0x80, 0x00]).unwrap(); // flags
                    res_buf.extend_from_slice(&buf[12..16]).unwrap(); // ciaddr
                    res_buf.extend_from_slice(&[192, 254, 0, 12]).unwrap(); // yiaddr
                    res_buf.extend_from_slice(&[0; 4]).unwrap(); // siaddr
                    res_buf.extend_from_slice(&buf[24..28]).unwrap(); // giaddr
                    res_buf.extend_from_slice(&buf[28..44]).unwrap(); // chaddr
                    res_buf.extend_from_slice(&[0; 192]).unwrap(); // sname/file
                    res_buf.extend_from_slice(&[99, 130, 83, 99]).unwrap(); // magic cookie

                    res_buf
                        .extend_from_slice(&[
                            53,
                            1,
                            match msg_type {
                                DhcpMsgType::DISCOVER => 2,
                                DhcpMsgType::REQUEST => 5,
                                _ => unreachable!(),
                            },
                        ])
                        .unwrap(); // opt message type
                    unwrap(
                        write_dhcp_opts(
                            &mut res_buf,
                            match msg_type {
                                DhcpMsgType::DISCOVER => {
                                    unwrap(v4_options!(msg; ParameterRequestList))
                                        .await
                                        .unwrap_or(&[])
                                }
                                DhcpMsgType::REQUEST => &[1, 3, 51, 6, 54],
                                _ => unreachable!(),
                            },
                        )
                        .await,
                    )
                    .await;
                    res_buf.push(255).unwrap(); // end option

                    for o in unwrap(unwrap(dhcpv4::Message::new(&res_buf)).await.options()).await {
                        info!(
                            "{:?}",
                            match o {
                                Err(_) => break,
                                Ok(o) => o,
                            }
                        );
                        Timer::after_secs(0).await;
                    }

                    unwrap(
                        socket
                            .send_to(
                                &res_buf,
                                UdpMetadata {
                                    endpoint: IpEndpoint::new(
                                        Ipv4Addr::new(255, 255, 255, 255).into(),
                                        68,
                                    ),
                                    local_address: Some(Ipv4Addr::new(192, 254, 0, 2).into()),
                                    meta: Default::default(),
                                },
                            )
                            .await,
                    )
                    .await;
                    info!("offer sent");
                    Timer::after_secs(0).await;
                }
                _ => {}
            }

            // unwrap(socket.send_to(&buf[..n], ep).await).await;
        }
    }

    // unwrap(spawner.spawn(socket::listen_task(stack, game::Team::Zero, 80))).await;
    // unwrap(spawner.spawn(socket::listen_task(stack, game::Team::One, 81))).await;
}

#[cfg(feature = "wifi-connect")]
#[derive(serde::Deserialize)]
struct WifiConnectConf<'a> {
    name: &'a str,
    password: &'a str,
    ip: Option<Ipv4Addr>,
}

async fn write_dhcp_opts<const N: usize>(buf: &mut Vec<u8, N>, op_codes: &[u8]) -> Result<(), ()> {
    for o in op_codes {
        let (opt_len, opt): (u8, &[u8]) = match o {
            1 => (4, &[255, 255, 255, 0]), // DhcpOption::SubnetMask(&dhcpv4::Addr([255, 255, 255, 0])),
            2 => (4, &3600_i32.to_be_bytes()), // DhcpOption::TimeOffset(3600),
            3 => (4, &[192, 254, 0, 2]),   // DhcpOption::Router(&[dhcpv4::Addr([192, 254, 0, 2])]),
            6 => (4, &[0, 0, 0, 0]), // DhcpOption::DomainNameServer(&[dhcpv4::Addr([0, 0, 0, 0])]),
            12 => (4, b"blue"),      // DhcpOption::HostName(b"blue"),
            15 => (11, b"LocalDomain"), // DhcpOption::DomainName(b"LocalDomain"),
            26 => (2, &1514_u16.to_be_bytes()), // DhcpOption::Unknown(26, &[0x5, 0xEA]), // mtu
            28 => (4, &[192, 254, 0, 255]), // DhcpOption::Unknown(28, &[192, 254, 0, 255]), // broadcast
            51 => (4, &700_u32.to_be_bytes()), // DhcpOption::AddressLeaseTime(700),
            54 => (4, &[192, 254, 0, 2]), // DhcpOption::ServerIdentifier(&dhcpv4::Addr([192, 254, 0, 2])),
            58 => (4, &500_u32.to_be_bytes()), // DhcpOption::Unknown(58, &[0, 0, 0x1, 0xF4]), // renewal time = 500s
            59 => (4, &600_u32.to_be_bytes()), // DhcpOption::Unknown(59, &[0, 0, 0x2, 0x58]), // rebinding time = 600s
            _ => {
                info!("Unhandled requested option : {}", o);
                Timer::after_secs(0).await;
                continue;
            }
        };
        buf.push(*o).map_err(|_| ())?;
        buf.push(opt_len).map_err(|_| ())?;
        buf.extend_from_slice(opt)?;
    }
    Ok(())
}
